import torch
import torch.nn as nn
import logging
from tqdm import tqdm   
import torch.nn.functional as F
import torch.optim as optim
import copy
from abc import ABC


logging.basicConfig(filename='model_training.log', level=logging.INFO,
                    format='%(asctime)s - %(levelname)s - %(message)s')


class BaseModel(ABC):
    def __init__(self, model, learning_rate=3e-4):
        self.device = 'cuda' if torch.cuda.is_available() else 'cpu'
        self.model = model.to(torch.float64).to(self.device)
        self.criterion = nn.CrossEntropyLoss()
        self.optimizer = optim.AdamW(self.model.parameters(), lr=learning_rate)
        logging.info(f"Initialised BaseModel with learning rate: {learning_rate}")
   

    def fit(self, train_loader, test_loader, num_epochs=50, debug=True):
        best_model_wts = copy.deepcopy(self.model.state_dict())
        best_acc = 0.0
        print("Model initialised. Begin training sequence...")
    
        for epoch in tqdm(range(num_epochs), desc="Model Training Progress"):
            train_loss, softmax_outputs = self._train_epoch(train_loader, debug)
            test_acc = self.evaluate(test_loader)
            
            log_message = f"Epoch {epoch+1}/{num_epochs} - Loss: {train_loss:.4f}, Test Accuracy: {test_acc:.2f}%"
            if debug and softmax_outputs:
                log_message += f"\nSoftmax outputs: {softmax_outputs}"
            logging.info(log_message)
    
            if test_acc > best_acc:
                best_acc = test_acc
                best_model_wts = copy.deepcopy(self.model.state_dict())
                torch.save(best_model_wts, 'best_model.pth')
                logging.info(f"New best model saved with accuracy: {best_acc:.2f}%")
    
        logging.info(f"Model training completed. Best test accuracy: {best_acc:.2f}%")
        print(f"Training and Eval Complete. best_accuracy: {best_acc:.2f}%. Please check model_training.log for additional info.")
        self.model.load_state_dict(best_model_wts)
        return best_acc
    
    def _train_epoch(self, loader, debug=True):
        self.model.train()
        epoch_loss = 0.0
        softmax_outputs = []
        for i, (X_train, y_train) in enumerate(loader):
            X_train = X_train.to(self.device)
            y_train = y_train.to(self.device).long() 
    
            self.optimizer.zero_grad()
            y_preds = self.model(X_train)
            loss = self.criterion(y_preds, y_train)
            loss.backward()
            self.optimizer.step()
    
            epoch_loss += loss.item()
    
            if i % 10 == 0 and debug:
                softmax_outputs.append(self._debug_output(i, loss, y_preds, y_train))
    
        return epoch_loss / len(loader), softmax_outputs
    
    def _debug_output(self, batch_index, loss, y_preds, y_train):
        probabilities = F.softmax(y_preds, dim=1)
        num_samples_to_log = min(5, y_preds.shape[0])
        output = []
        for j in range(num_samples_to_log):
            output.append({
                'Sample': j,
                'Probabilities': probabilities[j].tolist(),
                'True class': y_train[j].item()
            })
        return output
    @torch.no_grad()
    def evaluate(self, dataloader):
        self.model.eval()
        correct = 0
        total = 0
        for X, y_true in dataloader:
            X = X.to(self.device)
            y_true = y_true.to(self.device)
            outputs = self.model(X)
            _, predicted = torch.max(outputs.data, 1)
            total += y_true.size(0)
            correct += (predicted == y_true).sum().item()
        return 100 * correct / total



class SimpleModel(nn.Module):
    def __init__(self, input_size, hidden_size, output_size):
        super(SimpleModel, self).__init__()
        logging.info(f"Initializing SimpleModel - input_size: {input_size}, hidden_size: {hidden_size}, output_size: {output_size}")
        self.fc1 = nn.Linear(input_size, hidden_size)
        self.fc2 = nn.Linear(hidden_size, hidden_size)
        self.fc3 = nn.Linear(hidden_size, hidden_size)
        self.fc4 = nn.Linear(hidden_size, output_size)

    def forward(self, x):
        x = x.view(x.size(0), -1)
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = F.relu(self.fc3(x))
        x = self.fc4(x)
        return x
    

class IcarusModel(BaseModel):
    def __init__(self, input_size, hidden_size, output_size, learning_rate=3e-4):
        logging.info(f"Initializing IcarusModel - input_size: {input_size}, hidden_size: {hidden_size}, output_size: {output_size}, learning_rate: {learning_rate}")
        model = SimpleModel(input_size, hidden_size, output_size)
        super().__init__(model, learning_rate)


#Mooseherder can generate 125 sims per min 
from pathlib import Path
import warnings
import numpy as np
import matplotlib.pyplot as plt
from skopt.space import Space
from skopt.sampler import Sobol
from mooseherder import (MooseHerd,
                         MooseRunner,
                         InputModifier,
                         DirectoryManager,
                         MooseConfig)



warnings.filterwarnings("ignore", category=UserWarning, module="skopt.sampler.sobol")

N_SAMPLES = 100
np.random.seed(123)

class SampleGenerator:
    '''generates perturbed parameter samples. using quasirandom (sobol) sampling within stratified samples to ensure dataset balance and 
       and sufficient coverage of the continous sample space, as specified by the perturbation factor,  and the tolerance level, t. both hyperparamterters can be altered by user 
    '''
    def __init__(self, n_samples, thermal_cond_base, heat_flux_base, p_factor=0.7, tolerance=0.001):
        self.n_samples = n_samples
        self.thermal_cond_base = thermal_cond_base
        self.heat_flux_base = heat_flux_base
        self.p_factor = p_factor
        self.tolerance = tolerance
        self.thermal_c_samples, self.heat_flux_samples = self.generate_stratified_samples()
        self.classified_samples = self.classify_samples()

    def generate_stratified_samples(self):
        samples_per_class = self.n_samples // 4
        
        # define the ranges for each parameter
        tc_min = self.thermal_cond_base * (1 - self.p_factor)
        tc_max = self.thermal_cond_base * (1 + self.p_factor)
        hf_min = self.heat_flux_base * (1 - self.p_factor)
        hf_max = self.heat_flux_base * (1 + self.p_factor)
        
        # define the boundaries for "unperturbed" regions
        tc_low = self.thermal_cond_base * (1 - self.tolerance)
        tc_high = self.thermal_cond_base * (1 + self.tolerance)
        hf_low = self.heat_flux_base * (1 - self.tolerance)
        hf_high = self.heat_flux_base * (1 + self.tolerance)
        
        # Generate samples for each class
        class_0 = self.sobol_samples(tc_low, tc_high, hf_low, hf_high, samples_per_class)
        class_1 = self.sobol_samples(tc_low, tc_high, hf_min, hf_low, samples_per_class // 2) + \
                  self.sobol_samples(tc_low, tc_high, hf_high, hf_max, samples_per_class // 2)
        class_2 = self.sobol_samples(tc_min, tc_low, hf_low, hf_high, samples_per_class // 2) + \
                  self.sobol_samples(tc_high, tc_max, hf_low, hf_high, samples_per_class // 2)
        class_3 = self.sobol_samples(tc_min, tc_low, hf_min, hf_low, samples_per_class // 4) + \
                  self.sobol_samples(tc_min, tc_low, hf_high, hf_max, samples_per_class // 4) + \
                  self.sobol_samples(tc_high, tc_max, hf_min, hf_low, samples_per_class // 4) + \
                  self.sobol_samples(tc_high, tc_max, hf_high, hf_max, samples_per_class // 4)
        
        # Combine all samples
        all_samples = class_0 + class_1 + class_2 + class_3
        np.random.shuffle(all_samples)
        
        thermal_cond_vals = [sample[0] for sample in all_samples]
        heat_flux_vals = [sample[1] for sample in all_samples]
        
        return thermal_cond_vals, heat_flux_vals

    def sobol_samples(self, tc_min, tc_max, hf_min, hf_max, n_samples):
        space = Space([(tc_min, tc_max), (hf_min, hf_max)])
        samples = Sobol().generate(space.dimensions, n_samples)
        return [(sample[0], sample[1]) for sample in samples]

    def classify_samples(self):
        classified_samples = []
        for thermal_cond, heat_flux in zip(self.thermal_c_samples, self.heat_flux_samples):
            tc_perturbed = abs(thermal_cond - self.thermal_cond_base) > self.thermal_cond_base * self.tolerance
            hf_perturbed = abs(heat_flux - self.heat_flux_base) > self.heat_flux_base * self.tolerance
            
            class_label = tc_perturbed * 2 + hf_perturbed
            classified_samples.append((thermal_cond, heat_flux, class_label))
        
        return classified_samples

    def get_class_distribution(self):
        class_counts = [0, 0, 0, 0]
        for _, _, label in self.classified_samples:
            class_counts[label] += 1
        return class_counts

    def print_sample_info(self, num_samples=10):
        print(f"First {num_samples} samples:")
        for i, (tc, hf, cl) in enumerate(self.classified_samples[:num_samples]):
            print(f"Sample {i+1}: Thermal Conductivity = {tc:.2f}, Heat Flux = {hf:.2f}, Class = {cl}")

        class_counts = self.get_class_distribution()
        print("\nClass distribution:")
        for i, count in enumerate(class_counts):
            print(f"Class {i}: {count} samples")

    def plot_searchspace(self):
        fig, ax = plt.subplots()
        plt.plot(self.thermal_c_samples, self.heat_flux_samples, 'bo', label='samples')
        plt.plot(self.thermal_c_samples, self.heat_flux_samples, 'bo', markersize=5, alpha=0.2)
        ax.set_xlabel("Thermal Conductivity")
        ax.set_xlim([self.thermal_cond_base * (1-self.p_factor), self.thermal_cond_base * (1+self.p_factor)])
        ax.set_ylabel("Heat Flux")
        ax.set_ylim([self.heat_flux_base * (1-self.p_factor), self.heat_flux_base * (1+self.p_factor)])
        plt.title('Stratified Sobol Samples')
        plt.show()




from pathlib import Path
import numpy as np
from mooseherder import ExodusReader
import json
import numpy as np
from sklearn.model_selection import train_test_split
import matplotlib.pyplot as plt
import os
import re


thermal_plate_example = 'ground_truth_thermal_sim_out.e'
thermal_monoblock_example = 'monoblock_out.e'



def create_tensor(exodus_file):
    exodus_reader = ExodusReader(exodus_file)
    read_config = exodus_reader.get_read_config()
    read_config.node_vars = np.array([('temperature')])
    sim_data = exodus_reader.get_all_node_vars()
    temperature_tensor = np.array([value for value in sim_data.values()], dtype=np.float64)
    return temperature_tensor[:,:,1].T

BASE_DIR = Path.cwd()
OUTPUT_FILE = 'thermal_monoblock_data.npz'
GROUND_TRUTH_TENSOR = create_tensor(Path(BASE_DIR, thermal_monoblock_example))




def extract_number(filename):
    # Extract the number after the last dash and before _out.e
    match = re.search(r'sim-1-(\d+)_out\.e$', filename)
    if match:
        return int(match.group(1))
    return 0  # Return 0 if no match found

def preprocess_data(base_dir, output_file, test_size=0.2, random_state=42,):
    dataset_dir = Path(Path.cwd(), 'dataset/')
    working_dirs = [dir for dir in dataset_dir.iterdir() if dir.is_dir()]
    temperature_data = []
    y_data = []
    
    for working_dir in working_dirs:
        exodus_files = sorted(list(working_dir.glob('*.e')))
        sweep_vars_files= sorted(list(working_dir.glob('sweep-vars-*.json')))

    exodus_files_sorted = sorted(exodus_files, key=lambda x: extract_number(os.path.basename(x)))


    for exodus_file in exodus_files_sorted:
                temperature_tensor = create_tensor(exodus_file)
                temperature_data.append(GROUND_TRUTH_TENSOR - temperature_tensor)

    for sweep_vars_file in sweep_vars_files:
        y_data.extend(calculate_y_data(sweep_vars_file))

    temperature_data = np.array(temperature_data)
    max_abs_diff = np.max(np.abs(temperature_data))
    temperature_data_normalized = temperature_data / max_abs_diff
    y_data = np.array(y_data)


    # Split the data into train and test sets
    temperature_train, temperature_test, y_train, y_test = train_test_split(
        temperature_data, y_data, test_size=test_size, random_state=random_state, shuffle=False
    )


    np.savez(output_file, 
             temperature_train=temperature_train, 
             temperature_test=temperature_test,
             y_train=y_train, 
             y_test=y_test, 
             ground_truth=GROUND_TRUTH_TENSOR,
             max_abs_diff=max_abs_diff,
             ground_truth_thermal_c=384.0,
             ground_truth_surface_heat_flux=10.0e7)

# def create_tensor(exodus_file_path, scale_factor=1):
#     reader = ExodusReader(exodus_file_path)
#     coords, _ = reader.get_coords()
#     coords_2d = coords[:, :2]

#     temp_data = reader.get_node_vars(np.array(['temperature']))
#     if temp_data is None or 'temperature' not in temp_data:
#         raise ValueError("Temperature data not found in the exodus file")

#     temp_values = temp_data['temperature'][:, -1]

#     coord_temp_dict = {tuple(coord): temp for coord, temp in zip(coords_2d, temp_values)}

#     x = sorted(set(coord[0] for coord in coord_temp_dict.keys()))
#     y = sorted(set(coord[1] for coord in coord_temp_dict.keys()))

#     temp_grid = np.zeros((len(y), len(x)))
#     for (x_coord, y_coord), temp in coord_temp_dict.items():
#         i = y.index(y_coord)
#         j = x.index(x_coord)
#         temp_grid[i, j] = temp

#     # Create interpolation function
#     interp_func = RegularGridInterpolator((y, x), temp_grid)

#     # Create new coordinate arrays with higher resolution
#     x_new = np.linspace(min(x), max(x), len(x) * scale_factor)
#     y_new = np.linspace(min(y), max(y), len(y) * scale_factor)

#     # Create a meshgrid of new coordinates
#     xx_new, yy_new = np.meshgrid(x_new, y_new)
#     points_new = np.column_stack([yy_new.ravel(), xx_new.ravel()])

#     # Interpolate data
#     interpolated_data = interp_func(points_new).reshape(len(y_new), len(x_new))

#     return interpolated_data



def show_field(interpolated_data):
    plt.figure(figsize=(12, 5))
    plt.subplot(122)
    plt.imshow(interpolated_data, cmap='coolwarm', aspect='auto')
    plt.title("Interpolated Data")
    plt.colorbar()

    plt.tight_layout()
    plt.show()



def calculate_y_data(perturbed_data_file, thermal_cond_base=384.0, heat_flux_base=500000.0, tolerance=0.015):
    y_data = []

    with open(perturbed_data_file, 'r') as file:
        perturbed_data = json.load(file)

    for data_list in perturbed_data:
        data = data_list[0]  # Access the first (and only) dictionary in each inner list
        thermal_cond = data["cuThermCond"]
        heat_flux = data["surfHeatFlux"]

        # Check if parameters are perturbed based on tolerance
        tc_perturbed = abs(thermal_cond - thermal_cond_base) > thermal_cond_base * tolerance
        hf_perturbed = abs(heat_flux - heat_flux_base) > heat_flux_base * tolerance
        
        # Classify the sample
        if not tc_perturbed and not hf_perturbed:
            y_data.append(0)  # both unchanged
        elif not tc_perturbed and hf_perturbed:
            y_data.append(1)  # only heat flux perturbed
        elif tc_perturbed and not hf_perturbed:
            y_data.append(2)  # only thermal conductivity perturbed
        else:
            y_data.append(3)  # both perturbed

    return y_data

        

preprocess_data(BASE_DIR, OUTPUT_FILE)




import os
import numpy as np

def load_2d_thermal_dataset():
    """Load the built-in thermal model dataset."""
    file_path = os.path.join(os.path.dirname(__file__), 'data', 'thermal_model_data.npz')
    return dict(np.load(file_path, allow_pickle=True))

def load_monoblock_thermal_dataset():
    """Load the built-in thermal model dataset."""
    file_path = os.path.join(os.path.dirname(__file__), 'data', 'thermal_monoblock_data.npz')
    return dict(np.load(file_path, allow_pickle=True))


import numpy as np
import torch
from torch.utils.data import Dataset


class ThermalModelDataset(Dataset):
    def __init__(self, data_source, train=True):
        self.train = train

        if isinstance(data_source, str):
            # If data_source is a string, treat it as a file path
            self.__dataset_npz = np.load(data_source, allow_pickle=True)
        elif isinstance(data_source, dict):
            # If data_source is a dict, assume it's already loaded data
            self.__dataset_npz = data_source
        else:
            raise ValueError("data_source must be either a file path or a pre-loaded dataset dictionary")
       
       
        self.temperature_train = torch.from_numpy(self.__dataset_npz['temperature_train']).to(torch.float64)
        self.temperature_test = torch.from_numpy(self.__dataset_npz['temperature_test']).to(torch.float64)
        self.ground_truth = torch.from_numpy(self.__dataset_npz['ground_truth']).to(torch.float64)
        
        self.y_train = torch.from_numpy(self.__dataset_npz['y_train']).long()
        self.y_test = torch.from_numpy(self.__dataset_npz['y_test']).long()

        # self.max_abs_diff = self.__dataset_npz['max_abs_diff'].item()
        # self.ground_truth_thermal_c = self.__dataset_npz['ground_truth_thermal_c'].item()
        # self.ground_truth_surface_heat_flux = self.__dataset_npz['ground_truth_surface_heat_flux'].item()

        self.category = {0:'unperturbed', 1: 'heat_flux perturbed', 2:'thermal_cond_perturbed', 3:'both perturbed'}
        
        # self.temperature_train = F.normalize(self.temperature_train, dim=0)


        # self.temperature_test = F.normalize(self.temperature_test, dim=0)

        # self.y_train = self.y_train.view(-1,4,1)
        # self.y_test = self.y_test.view(-1,4,1)


    def __len__(self):
        if self.train:
            return len(self.temperature_train)
        else:
            return len(self.temperature_test)

    def __getitem__(self, index):
        if self.train:
            feature, label = self.temperature_train[index], self.y_train[index]
        else:
            feature, label = self.temperature_test[index], self.y_test[index]
        
        return feature, label
    


    # def denormalize(self, normalized_values):
    #     """
    #     Denormalize the predictions or true values.
    #     """
    #     denorm_thermal_c = normalized_values[:, 0] * (0.5 * self.ground_truth_thermal_c)
    #     denorm_heat_flux = normalized_values[:, 1] * (0.5 * self.ground_truth_surface_heat_flux)
    #     return torch.stack((denorm_thermal_c, denorm_heat_flux), dim=1)

    # def denormalize_temperature(self, normalized_temp):
    #     """
    #     Denormalize the temperature data.
    #     """
    #     return normalized_temp * self.max_abs_diff
    
    
    def __repr__(self) -> str:
        return f"""Thermal Model Simulation Validation/Verification Dataset:
    --------
    Training set: {len(self.temperature_train)} simulated temperature fields
    Testing set:  {len(self.temperature_test)} simulated temperature fields
    Input X data shape:           {self.temperature_train[0].shape}
    Input Y data shape:           {self.y_train[0].shape}
    Output classes:        4
    --------"""

    
        








from pathlib import Path
from mooseherder import MooseConfig


def main() -> None:
    """main: create moose config json
    """
    config = {'main_path': Path.home()/ 'moose',
            'app_path': Path.home() / 'proteus',
            'app_name': 'proteus-opt'}

    moose_config = MooseConfig(config)

    save_path = Path.cwd() / 'moose-config.json'
    moose_config.save_config(save_path)


if __name__ == "__main__":
    main()


# -*- coding: utf-8 -*-
"""
icarus
"""

from icarus.sample_generator import SampleGenerator
from icarus.dataloader import ThermalModelDataset
from icarus.model import IcarusModel
from icarus.datasets import load_2d_thermal_dataset, load_monoblock_thermal_dataset


__all__ = ["thermalmodeldataset",
            "samplegenerator",
            "icarusmodel"]